@@ -24,6 +24,9 @@
 #include "gstd3d11overlaycompositor.h"
 #include "gstd3d11pluginutils.h"
 #include <wrl.h>
+#include <memory>
+#include <vector>
+#include <algorithm>
 
 GST_DEBUG_CATEGORY_EXTERN (gst_d3d11_overlay_compositor_debug);
 #define GST_CAT_DEFAULT gst_d3d11_overlay_compositor_debug
@@ -44,37 +47,25 @@ typedef struct
   } texture;
 } VertexData;
 
-static const gchar templ_pixel_shader[] =
-    "Texture2D shaderTexture;\n"
-    "SamplerState samplerState;\n"
-    "struct PS_INPUT\n"
-    "{\n"
-    "  float4 Position: SV_POSITION;\n"
-    "  float2 Texture: TEXCOORD;\n"
-    "};\n"
-    "float4 main(PS_INPUT input): SV_TARGET\n"
-    "{\n"
-    "  return shaderTexture.Sample(samplerState, input.Texture);\n"
-    "}\n";
+struct GstD3D11CompositionOverlay
+{
+  ~GstD3D11CompositionOverlay ()
+  {
+    if (overlay_rect)
+      gst_video_overlay_rectangle_unref (overlay_rect);
+    if (d3d11_buffer)
+      gst_buffer_unref (d3d11_buffer);
+  }
 
-static const gchar templ_vertex_shader[] =
-    "struct VS_INPUT\n"
-    "{\n"
-    "  float4 Position : POSITION;\n"
-    "  float2 Texture : TEXCOORD;\n"
-    "};\n"
-    "\n"
-    "struct VS_OUTPUT\n"
-    "{\n"
-    "  float4 Position: SV_POSITION;\n"
-    "  float2 Texture: TEXCOORD;\n"
-    "};\n"
-    "\n"
-    "VS_OUTPUT main(VS_INPUT input)\n"
-    "{\n"
-    "  return input;\n"
-    "}\n";
-/* *INDENT-ON* */
+  GstVideoOverlayRectangle *overlay_rect = nullptr;
+  ComPtr<ID3D11Texture2D> texture;
+  ComPtr<ID3D11ShaderResourceView> srv;
+  ComPtr<ID3D11Buffer> vertex_buffer;
+  gboolean premul_alpha = FALSE;
+  GstBuffer *d3d11_buffer = nullptr;
+};
+
+typedef std::shared_ptr<GstD3D11CompositionOverlay> GstD3D11CompositionOverlayPtr;
 
 struct _GstD3D11OverlayCompositorPrivate
 {
@@ -82,31 +73,24 @@ struct _GstD3D11OverlayCompositorPrivate
 
   D3D11_VIEWPORT viewport;
 
-  ID3D11PixelShader *ps;
-  ID3D11VertexShader *vs;
-  ID3D11InputLayout *layout;
-  ID3D11SamplerState *sampler;
-  ID3D11BlendState *blend;
-  ID3D11Buffer *index_buffer;
+  ComPtr<ID3D11PixelShader> ps;
+  ComPtr<ID3D11PixelShader> premul_ps;
+  ComPtr<ID3D11VertexShader> vs;
+  ComPtr<ID3D11InputLayout> layout;
+  ComPtr<ID3D11SamplerState> sampler;
+  ComPtr<ID3D11BlendState> blend;
+  ComPtr<ID3D11Buffer> index_buffer;
+  ComPtr<ID3D11RasterizerState> rs;
+  std::vector<GstVideoOverlayRectangle *> rects_to_upload;
 
-  /* GstD3D11CompositionOverlay */
-  GList *overlays;
+  std::vector<GstD3D11CompositionOverlayPtr> overlays;
 };
+/* *INDENT-ON* */
 
-typedef struct
-{
-  GstVideoOverlayRectangle *overlay_rect;
-  ID3D11Texture2D *texture;
-  ID3D11ShaderResourceView *srv;
-  ID3D11Buffer *vertex_buffer;
-} GstD3D11CompositionOverlay;
-
-static void gst_d3d11_overlay_compositor_dispose (GObject * object);
-static void
-gst_d3d11_overlay_compositor_free_overlays (GstD3D11OverlayCompositor * self);
+static void gst_d3d11_overlay_compositor_finalize (GObject * object);
 
 #define gst_d3d11_overlay_compositor_parent_class parent_class
-G_DEFINE_TYPE_WITH_PRIVATE (GstD3D11OverlayCompositor,
+G_DEFINE_TYPE (GstD3D11OverlayCompositor,
     gst_d3d11_overlay_compositor, GST_TYPE_OBJECT);
 
 static void
@@ -114,50 +98,39 @@ gst_d3d11_overlay_compositor_class_init (GstD3D11OverlayCompositorClass * klass)
 {
   GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
-  object_class->dispose = gst_d3d11_overlay_compositor_dispose;
+  object_class->finalize = gst_d3d11_overlay_compositor_finalize;
 }
 
 static void
 gst_d3d11_overlay_compositor_init (GstD3D11OverlayCompositor * self)
 {
-  self->priv = (GstD3D11OverlayCompositorPrivate *)
-      gst_d3d11_overlay_compositor_get_instance_private (self);
+  self->priv = new GstD3D11OverlayCompositorPrivate ();
 }
 
 static void
-gst_d3d11_overlay_compositor_dispose (GObject * object)
+gst_d3d11_overlay_compositor_finalize (GObject * object)
 {
   GstD3D11OverlayCompositor *self = GST_D3D11_OVERLAY_COMPOSITOR (object);
-  GstD3D11OverlayCompositorPrivate *priv = self->priv;
 
-  gst_d3d11_overlay_compositor_free_overlays (self);
-
-  GST_D3D11_CLEAR_COM (priv->ps);
-  GST_D3D11_CLEAR_COM (priv->vs);
-  GST_D3D11_CLEAR_COM (priv->layout);
-  GST_D3D11_CLEAR_COM (priv->sampler);
-  GST_D3D11_CLEAR_COM (priv->blend);
-  GST_D3D11_CLEAR_COM (priv->index_buffer);
+  delete self->priv;
 
   gst_clear_object (&self->device);
 
-  G_OBJECT_CLASS (parent_class)->dispose (object);
+  G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
-static GstD3D11CompositionOverlay *
+static GstD3D11CompositionOverlayPtr
 gst_d3d11_composition_overlay_new (GstD3D11OverlayCompositor * self,
     GstVideoOverlayRectangle * overlay_rect)
 {
   GstD3D11OverlayCompositorPrivate *priv = self->priv;
-  GstD3D11CompositionOverlay *overlay = nullptr;
   gint x, y;
   guint width, height;
-  D3D11_SUBRESOURCE_DATA subresource_data;
+  D3D11_SUBRESOURCE_DATA subresource;
   D3D11_TEXTURE2D_DESC texture_desc;
   D3D11_SHADER_RESOURCE_VIEW_DESC srv_desc;
   D3D11_BUFFER_DESC buffer_desc;
-  D3D11_MAPPED_SUBRESOURCE map;
-  VertexData *vertex_data;
+  VertexData vertex_data[4];
   GstBuffer *buf;
   GstVideoMeta *vmeta;
   GstMapInfo info;
@@ -165,21 +138,23 @@ gst_d3d11_composition_overlay_new (GstD3D11OverlayCompositor * self,
   gint stride;
   HRESULT hr;
   ID3D11Device *device_handle;
-  ID3D11DeviceContext *context_handle;
   GstD3D11Device *device = self->device;
   FLOAT x1, y1, x2, y2;
   gdouble val;
   ComPtr < ID3D11Texture2D > texture;
   ComPtr < ID3D11ShaderResourceView > srv;
   ComPtr < ID3D11Buffer > vertex_buffer;
+  GstVideoOverlayFormatFlags flags;
+  gboolean premul_alpha = FALSE;
+  GstMemory *mem;
+  gboolean is_d3d11 = FALSE;
 
-  memset (&subresource_data, 0, sizeof (subresource_data));
+  memset (&subresource, 0, sizeof (subresource));
   memset (&texture_desc, 0, sizeof (texture_desc));
   memset (&srv_desc, 0, sizeof (srv_desc));
   memset (&buffer_desc, 0, sizeof (buffer_desc));
 
   device_handle = gst_d3d11_device_get_device_handle (device);
-  context_handle = gst_d3d11_device_get_device_context_handle (device);
 
   if (!gst_video_overlay_rectangle_get_render_rectangle (overlay_rect, &x, &y,
           &width, &height)) {
@@ -187,82 +162,81 @@ gst_d3d11_composition_overlay_new (GstD3D11OverlayCompositor * self,
     return nullptr;
   }
 
+  flags = gst_video_overlay_rectangle_get_flags (overlay_rect);
+  if ((flags & GST_VIDEO_OVERLAY_FORMAT_FLAG_PREMULTIPLIED_ALPHA) != 0) {
+    premul_alpha = TRUE;
+    flags = GST_VIDEO_OVERLAY_FORMAT_FLAG_PREMULTIPLIED_ALPHA;
+  } else {
+    flags = GST_VIDEO_OVERLAY_FORMAT_FLAG_NONE;
+  }
+
   buf = gst_video_overlay_rectangle_get_pixels_unscaled_argb (overlay_rect,
-      GST_VIDEO_OVERLAY_FORMAT_FLAG_NONE);
+      flags);
   if (!buf) {
     GST_ERROR_OBJECT (self, "Failed to get overlay buffer");
     return nullptr;
   }
 
-  vmeta = gst_buffer_get_video_meta (buf);
-  if (!vmeta) {
-    GST_ERROR_OBJECT (self, "Failed to get video meta");
-    return nullptr;
+  mem = gst_buffer_peek_memory (buf, 0);
+  if (gst_is_d3d11_memory (mem)) {
+    GstD3D11Memory *dmem = GST_D3D11_MEMORY_CAST (mem);
+    if (dmem->device == self->device) {
+      srv = gst_d3d11_memory_get_shader_resource_view (dmem, 0);
+      if (srv) {
+        texture = (ID3D11Texture2D *)
+            gst_d3d11_memory_get_resource_handle (dmem);
+        is_d3d11 = TRUE;
+      }
+    }
   }
 
-  if (!gst_video_meta_map (vmeta,
-          0, &info, (gpointer *) & data, &stride, GST_MAP_READ)) {
-    GST_ERROR_OBJECT (self, "Failed to map");
-    return nullptr;
+  if (!is_d3d11) {
+    vmeta = gst_buffer_get_video_meta (buf);
+    if (!vmeta) {
+      GST_ERROR_OBJECT (self, "Failed to get video meta");
+      return nullptr;
+    }
+
+    if (!gst_video_meta_map (vmeta,
+            0, &info, (gpointer *) & data, &stride, GST_MAP_READ)) {
+      GST_ERROR_OBJECT (self, "Failed to map");
+      return nullptr;
+    }
+
+    /* Do create texture and upload data at once, for create immutable texture */
+    subresource.pSysMem = data;
+    subresource.SysMemPitch = stride;
+
+    texture_desc.Width = vmeta->width;
+    texture_desc.Height = vmeta->height;
+    texture_desc.MipLevels = 1;
+    texture_desc.ArraySize = 1;
+    texture_desc.Format = DXGI_FORMAT_B8G8R8A8_UNORM;
+    texture_desc.SampleDesc.Count = 1;
+    texture_desc.SampleDesc.Quality = 0;
+    texture_desc.Usage = D3D11_USAGE_IMMUTABLE;
+    texture_desc.BindFlags = D3D11_BIND_SHADER_RESOURCE;
+    texture_desc.CPUAccessFlags = 0;
+
+    hr = device_handle->CreateTexture2D (&texture_desc, &subresource, &texture);
+    gst_video_meta_unmap (vmeta, 0, &info);
+
+    if (!gst_d3d11_result (hr, device)) {
+      GST_ERROR_OBJECT (self, "Failed to create texture");
+      return nullptr;
+    }
+
+    srv_desc.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE2D;
+    srv_desc.Texture2D.MipLevels = 1;
+
+    hr = device_handle->CreateShaderResourceView (texture.Get (), &srv_desc,
+        &srv);
+    if (!gst_d3d11_result (hr, device) || !srv) {
+      GST_ERROR_OBJECT (self, "Failed to create shader resource view");
+      return nullptr;
+    }
   }
 
-  /* Do create texture and upload data at once, for create immutable texture */
-  subresource_data.pSysMem = data;
-  subresource_data.SysMemPitch = stride;
-  subresource_data.SysMemSlicePitch = 0;
-
-  texture_desc.Width = vmeta->width;
-  texture_desc.Height = vmeta->height;
-  texture_desc.MipLevels = 1;
-  texture_desc.ArraySize = 1;
-  texture_desc.Format = DXGI_FORMAT_B8G8R8A8_UNORM;
-  texture_desc.SampleDesc.Count = 1;
-  texture_desc.SampleDesc.Quality = 0;
-  texture_desc.Usage = D3D11_USAGE_IMMUTABLE;
-  texture_desc.BindFlags = D3D11_BIND_SHADER_RESOURCE;
-  texture_desc.CPUAccessFlags = 0;
-
-  hr = device_handle->CreateTexture2D (&texture_desc,
-      &subresource_data, &texture);
-  gst_video_meta_unmap (vmeta, 0, &info);
-
-  if (!gst_d3d11_result (hr, device)) {
-    GST_ERROR_OBJECT (self, "Failed to create texture");
-    return nullptr;
-  }
-
-  srv_desc.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE2D;
-  srv_desc.Texture2D.MipLevels = 1;
-
-  hr = device_handle->CreateShaderResourceView (texture.Get (), &srv_desc,
-      &srv);
-  if (!gst_d3d11_result (hr, device) || !srv) {
-    GST_ERROR_OBJECT (self, "Failed to create shader resource view");
-    return nullptr;
-  }
-
-  buffer_desc.Usage = D3D11_USAGE_DYNAMIC;
-  buffer_desc.ByteWidth = sizeof (VertexData) * 4;
-  buffer_desc.BindFlags = D3D11_BIND_VERTEX_BUFFER;
-  buffer_desc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
-
-  hr = device_handle->CreateBuffer (&buffer_desc, nullptr, &vertex_buffer);
-  if (!gst_d3d11_result (hr, device)) {
-    GST_ERROR_OBJECT (self,
-        "Couldn't create vertex buffer, hr: 0x%x", (guint) hr);
-    return nullptr;
-  }
-
-  GstD3D11DeviceLockGuard lk (device);
-  hr = context_handle->Map (vertex_buffer.Get (),
-      0, D3D11_MAP_WRITE_DISCARD, 0, &map);
-
-  if (!gst_d3d11_result (hr, device)) {
-    GST_ERROR_OBJECT (self, "Couldn't map vertex buffer, hr: 0x%x", (guint) hr);
-    return nullptr;
-  }
-
-  vertex_data = (VertexData *) map.pData;
   /* bottom left */
   gst_util_fraction_to_double (x, GST_VIDEO_INFO_WIDTH (&priv->info), &val);
   x1 = (val * 2.0f) - 1.0f;
@@ -307,33 +281,33 @@ gst_d3d11_composition_overlay_new (GstD3D11OverlayCompositor * self,
   vertex_data[3].texture.u = 1.0f;
   vertex_data[3].texture.v = 1.0f;
 
-  context_handle->Unmap (vertex_buffer.Get (), 0);
+  buffer_desc.Usage = D3D11_USAGE_DYNAMIC;
+  buffer_desc.ByteWidth = sizeof (VertexData) * 4;
+  buffer_desc.BindFlags = D3D11_BIND_VERTEX_BUFFER;
+  buffer_desc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
 
-  overlay = g_new0 (GstD3D11CompositionOverlay, 1);
+  subresource.pSysMem = vertex_data;
+  subresource.SysMemPitch = sizeof (VertexData) * 4;
+
+  hr = device_handle->CreateBuffer (&buffer_desc, &subresource, &vertex_buffer);
+  if (!gst_d3d11_result (hr, device)) {
+    GST_ERROR_OBJECT (self,
+        "Couldn't create vertex buffer, hr: 0x%x", (guint) hr);
+    return nullptr;
+  }
+
+  auto overlay = std::make_shared < GstD3D11CompositionOverlay > ();
   overlay->overlay_rect = gst_video_overlay_rectangle_ref (overlay_rect);
-  overlay->texture = texture.Detach ();
-  overlay->srv = srv.Detach ();
-  overlay->vertex_buffer = vertex_buffer.Detach ();
+  overlay->texture = texture;
+  overlay->srv = srv;
+  overlay->vertex_buffer = vertex_buffer;
+  overlay->premul_alpha = premul_alpha;
+  if (is_d3d11)
+    overlay->d3d11_buffer = gst_buffer_ref (buf);
 
   return overlay;
 }
 
-static void
-gst_d3d11_composition_overlay_free (GstD3D11CompositionOverlay * overlay)
-{
-  if (!overlay)
-    return;
-
-  if (overlay->overlay_rect)
-    gst_video_overlay_rectangle_unref (overlay->overlay_rect);
-
-  GST_D3D11_CLEAR_COM (overlay->srv);
-  GST_D3D11_CLEAR_COM (overlay->texture);
-  GST_D3D11_CLEAR_COM (overlay->vertex_buffer);
-
-  g_free (overlay);
-}
-
 static gboolean
 gst_d3d11_overlay_compositor_setup_shader (GstD3D11OverlayCompositor * self)
 {
@@ -341,83 +315,66 @@ gst_d3d11_overlay_compositor_setup_shader (GstD3D11OverlayCompositor * self)
   GstVideoInfo *info = &priv->info;
   GstD3D11Device *device = self->device;
   HRESULT hr;
-  D3D11_SAMPLER_DESC sampler_desc;
-  D3D11_INPUT_ELEMENT_DESC input_desc[2];
   D3D11_BUFFER_DESC buffer_desc;
   D3D11_BLEND_DESC blend_desc;
-  D3D11_MAPPED_SUBRESOURCE map;
-  WORD *indices;
+  D3D11_SUBRESOURCE_DATA subresource;
+  const WORD indices[6] = { 0, 1, 2, 3, 0, 2 };
   ID3D11Device *device_handle;
-  ID3D11DeviceContext *context_handle;
   ComPtr < ID3D11PixelShader > ps;
+  ComPtr < ID3D11PixelShader > premul_ps;
   ComPtr < ID3D11VertexShader > vs;
   ComPtr < ID3D11InputLayout > layout;
   ComPtr < ID3D11SamplerState > sampler;
   ComPtr < ID3D11BlendState > blend;
   ComPtr < ID3D11Buffer > index_buffer;
+  ComPtr < ID3D11RasterizerState > rs;
 
-  memset (&sampler_desc, 0, sizeof (sampler_desc));
-  memset (input_desc, 0, sizeof (input_desc));
   memset (&buffer_desc, 0, sizeof (buffer_desc));
   memset (&blend_desc, 0, sizeof (blend_desc));
+  memset (&subresource, 0, sizeof (subresource));
 
   device_handle = gst_d3d11_device_get_device_handle (device);
-  context_handle = gst_d3d11_device_get_device_context_handle (device);
 
-  /* bilinear filtering */
-  sampler_desc.Filter = D3D11_FILTER_MIN_MAG_LINEAR_MIP_POINT;
-  sampler_desc.AddressU = D3D11_TEXTURE_ADDRESS_CLAMP;
-  sampler_desc.AddressV = D3D11_TEXTURE_ADDRESS_CLAMP;
-  sampler_desc.AddressW = D3D11_TEXTURE_ADDRESS_CLAMP;
-  sampler_desc.ComparisonFunc = D3D11_COMPARISON_ALWAYS;
-  sampler_desc.MinLOD = 0;
-  sampler_desc.MaxLOD = D3D11_FLOAT32_MAX;
-
-  hr = device_handle->CreateSamplerState (&sampler_desc, &sampler);
+  hr = gst_d3d11_device_get_sampler (device,
+      D3D11_FILTER_MIN_MAG_LINEAR_MIP_POINT, &sampler);
   if (!gst_d3d11_result (hr, device)) {
     GST_ERROR_OBJECT (self, "Couldn't create sampler state, hr: 0x%x",
         (guint) hr);
     return FALSE;
   }
 
-  hr = gst_d3d11_create_pixel_shader_simple (device,
-      templ_pixel_shader, "main", &ps);
+  hr = gst_d3d11_get_pixel_shader_sample (device, &ps);
   if (!gst_d3d11_result (hr, device)) {
     GST_ERROR_OBJECT (self, "Couldn't create pixel shader");
     return FALSE;
   }
 
-  input_desc[0].SemanticName = "POSITION";
-  input_desc[0].SemanticIndex = 0;
-  input_desc[0].Format = DXGI_FORMAT_R32G32B32_FLOAT;
-  input_desc[0].InputSlot = 0;
-  input_desc[0].AlignedByteOffset = D3D11_APPEND_ALIGNED_ELEMENT;
-  input_desc[0].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
-  input_desc[0].InstanceDataStepRate = 0;
-
-  input_desc[1].SemanticName = "TEXCOORD";
-  input_desc[1].SemanticIndex = 0;
-  input_desc[1].Format = DXGI_FORMAT_R32G32_FLOAT;
-  input_desc[1].InputSlot = 0;
-  input_desc[1].AlignedByteOffset = D3D11_APPEND_ALIGNED_ELEMENT;
-  input_desc[1].InputSlotClass = D3D11_INPUT_PER_VERTEX_DATA;
-  input_desc[1].InstanceDataStepRate = 0;
-
-  hr = gst_d3d11_create_vertex_shader_simple (device, templ_vertex_shader,
-      "main", input_desc, G_N_ELEMENTS (input_desc), &vs, &layout);
+  hr = gst_d3d11_get_pixel_shader_sample_premul (device, &premul_ps);
   if (!gst_d3d11_result (hr, device)) {
-    GST_ERROR_OBJECT (self, "Couldn't vertex pixel shader");
+    GST_ERROR_OBJECT (self, "Couldn't create premul pixel shader");
+    return FALSE;
+  }
+
+  hr = gst_d3d11_get_vertex_shader_coord (device, &vs, &layout);
+  if (!gst_d3d11_result (hr, device)) {
+    GST_ERROR_OBJECT (self, "Couldn't create vertex pixel shader");
+    return FALSE;
+  }
+
+  hr = gst_d3d11_device_get_rasterizer (device, &rs);
+  if (!gst_d3d11_result (hr, device)) {
+    GST_ERROR_OBJECT (self, "Couldn't get rasterizer state");
     return FALSE;
   }
 
   blend_desc.AlphaToCoverageEnable = FALSE;
   blend_desc.IndependentBlendEnable = FALSE;
   blend_desc.RenderTarget[0].BlendEnable = TRUE;
-  blend_desc.RenderTarget[0].SrcBlend = D3D11_BLEND_SRC_ALPHA;
+  blend_desc.RenderTarget[0].SrcBlend = D3D11_BLEND_ONE;
   blend_desc.RenderTarget[0].DestBlend = D3D11_BLEND_INV_SRC_ALPHA;
   blend_desc.RenderTarget[0].BlendOp = D3D11_BLEND_OP_ADD;
   blend_desc.RenderTarget[0].SrcBlendAlpha = D3D11_BLEND_ONE;
-  blend_desc.RenderTarget[0].DestBlendAlpha = D3D11_BLEND_ZERO;
+  blend_desc.RenderTarget[0].DestBlendAlpha = D3D11_BLEND_INV_SRC_ALPHA;
   blend_desc.RenderTarget[0].BlendOpAlpha = D3D11_BLEND_OP_ADD;
   blend_desc.RenderTarget[0].RenderTargetWriteMask =
       D3D11_COLOR_WRITE_ENABLE_ALL;
@@ -434,41 +391,24 @@ gst_d3d11_overlay_compositor_setup_shader (GstD3D11OverlayCompositor * self)
   buffer_desc.BindFlags = D3D11_BIND_INDEX_BUFFER;
   buffer_desc.CPUAccessFlags = D3D11_CPU_ACCESS_WRITE;
 
-  hr = device_handle->CreateBuffer (&buffer_desc, nullptr, &index_buffer);
+  subresource.pSysMem = indices;
+  subresource.SysMemPitch = sizeof (WORD) * 6;
+
+  hr = device_handle->CreateBuffer (&buffer_desc, &subresource, &index_buffer);
   if (!gst_d3d11_result (hr, device)) {
     GST_ERROR_OBJECT (self,
         "Couldn't create index buffer, hr: 0x%x", (guint) hr);
     return FALSE;
   }
 
-  GstD3D11DeviceLockGuard lk (device);
-  hr = context_handle->Map (index_buffer.Get (),
-      0, D3D11_MAP_WRITE_DISCARD, 0, &map);
-
-  if (!gst_d3d11_result (hr, device)) {
-    GST_ERROR_OBJECT (self, "Couldn't map index buffer, hr: 0x%x", (guint) hr);
-    return FALSE;
-  }
-
-  indices = (WORD *) map.pData;
-
-  /* clockwise indexing */
-  indices[0] = 0;               /* bottom left */
-  indices[1] = 1;               /* top left */
-  indices[2] = 2;               /* top right */
-
-  indices[3] = 3;               /* bottom right */
-  indices[4] = 0;               /* bottom left  */
-  indices[5] = 2;               /* top right */
-
-  context_handle->Unmap (index_buffer.Get (), 0);
-
-  priv->ps = ps.Detach ();
-  priv->vs = vs.Detach ();
-  priv->layout = layout.Detach ();
-  priv->sampler = sampler.Detach ();
-  priv->blend = blend.Detach ();
-  priv->index_buffer = index_buffer.Detach ();
+  priv->ps = ps;
+  priv->premul_ps = premul_ps;
+  priv->vs = vs;
+  priv->layout = layout;
+  priv->sampler = sampler;
+  priv->blend = blend;
+  priv->index_buffer = index_buffer;
+  priv->rs = rs;
 
   priv->viewport.TopLeftX = 0;
   priv->viewport.TopLeftY = 0;
@@ -506,39 +446,28 @@ gst_d3d11_overlay_compositor_new (GstD3D11Device * device,
   return self;
 }
 
-static void
-gst_d3d11_overlay_compositor_free_overlays (GstD3D11OverlayCompositor * self)
+static gboolean
+gst_d3d11_overlay_compositor_foreach_meta (GstBuffer * buffer, GstMeta ** meta,
+    GstD3D11OverlayCompositor * self)
 {
   GstD3D11OverlayCompositorPrivate *priv = self->priv;
+  GstVideoOverlayCompositionMeta *cmeta;
+  guint num_rect;
 
-  if (priv->overlays) {
-    g_list_free_full (priv->overlays,
-        (GDestroyNotify) gst_d3d11_composition_overlay_free);
+  if ((*meta)->info->api != GST_VIDEO_OVERLAY_COMPOSITION_META_API_TYPE)
+    return TRUE;
 
-    priv->overlays = nullptr;
+  cmeta = (GstVideoOverlayCompositionMeta *) (*meta);
+  if (!cmeta->overlay)
+    return TRUE;
+
+  num_rect = gst_video_overlay_composition_n_rectangles (cmeta->overlay);
+  for (guint i = 0; i < num_rect; i++) {
+    auto rect = gst_video_overlay_composition_get_rectangle (cmeta->overlay, i);
+    priv->rects_to_upload.push_back (rect);
   }
-}
 
-static gint
-find_in_compositor (const GstD3D11CompositionOverlay * overlay,
-    const GstVideoOverlayRectangle * rect)
-{
-  return !(overlay->overlay_rect == rect);
-}
-
-static gboolean
-is_in_video_overlay_composition (GstVideoOverlayComposition * voc,
-    GstD3D11CompositionOverlay * overlay)
-{
-  guint i;
-
-  for (i = 0; i < gst_video_overlay_composition_n_rectangles (voc); i++) {
-    GstVideoOverlayRectangle *rectangle =
-        gst_video_overlay_composition_get_rectangle (voc, i);
-    if (overlay->overlay_rect == rectangle)
-      return TRUE;
-  }
-  return FALSE;
+  return TRUE;
 }
 
 gboolean
@@ -546,61 +475,64 @@ gst_d3d11_overlay_compositor_upload (GstD3D11OverlayCompositor * compositor,
     GstBuffer * buf)
 {
   GstD3D11OverlayCompositorPrivate *priv;
-  GstVideoOverlayCompositionMeta *meta;
-  gint i, num_overlays;
-  GList *iter;
 
   g_return_val_if_fail (compositor != nullptr, FALSE);
   g_return_val_if_fail (GST_IS_BUFFER (buf), FALSE);
 
   priv = compositor->priv;
+  priv->rects_to_upload.clear ();
 
-  meta = gst_buffer_get_video_overlay_composition_meta (buf);
-  if (!meta) {
-    gst_d3d11_overlay_compositor_free_overlays (compositor);
+  gst_buffer_foreach_meta (buf,
+      (GstBufferForeachMetaFunc) gst_d3d11_overlay_compositor_foreach_meta,
+      compositor);
+
+  if (priv->rects_to_upload.empty ()) {
+    priv->overlays.clear ();
     return TRUE;
   }
 
-  num_overlays = gst_video_overlay_composition_n_rectangles (meta->overlay);
-  if (!num_overlays) {
-    gst_d3d11_overlay_compositor_free_overlays (compositor);
-    return TRUE;
-  }
+  GST_LOG_OBJECT (compositor, "Found %" G_GSIZE_FORMAT
+      " overlay rectangles, %" G_GSIZE_FORMAT " in current queue",
+      priv->rects_to_upload.size (), priv->overlays.size ());
 
-  GST_LOG_OBJECT (compositor, "Upload %d overlay rectangles", num_overlays);
-
-  /* Upload new overlay */
-  for (i = 0; i < num_overlays; i++) {
-    GstVideoOverlayRectangle *rectangle =
-        gst_video_overlay_composition_get_rectangle (meta->overlay, i);
-
-    if (!g_list_find_custom (priv->overlays,
-            rectangle, (GCompareFunc) find_in_compositor)) {
-      GstD3D11CompositionOverlay *overlay = nullptr;
-
-      overlay = gst_d3d11_composition_overlay_new (compositor, rectangle);
-
-      if (!overlay)
+  /* *INDENT-OFF* */
+  for (auto it : priv->rects_to_upload) {
+    if (std::find_if (priv->overlays.begin (), priv->overlays.end (),
+          [&] (const auto & overlay) -> bool {
+            return overlay->overlay_rect == it;
+           }) == priv->overlays.end ()) {
+      auto new_overlay = gst_d3d11_composition_overlay_new (compositor, it);
+      if (!new_overlay)
         return FALSE;
 
-      priv->overlays = g_list_append (priv->overlays, overlay);
+      priv->overlays.push_back (new_overlay);
     }
   }
+  /* *INDENT-ON* */
+
+  GST_LOG_OBJECT (compositor, "Overlay rectangles in queue after uploaded %"
+      G_GSIZE_FORMAT, priv->overlays.size ());
 
   /* Remove old overlay */
-  iter = priv->overlays;
-  while (iter) {
-    GstD3D11CompositionOverlay *overlay =
-        (GstD3D11CompositionOverlay *) iter->data;
-    GList *next = iter->next;
-
-    if (!is_in_video_overlay_composition (meta->overlay, overlay)) {
-      priv->overlays = g_list_delete_link (priv->overlays, iter);
-      gst_d3d11_composition_overlay_free (overlay);
+  /* *INDENT-OFF* */
+  auto it = priv->overlays.begin ();
+  while (it != priv->overlays.end ()) {
+    auto old_overlay = *it;
+    if (std::find_if (priv->rects_to_upload.begin (),
+          priv->rects_to_upload.end (), [&] (const auto & overlay) -> bool {
+            return overlay == old_overlay->overlay_rect;
+          }) == priv->rects_to_upload.end ()) {
+      GST_LOG_OBJECT (compositor, "Removing %p from queue",
+          old_overlay->overlay_rect);
+      it = priv->overlays.erase (it);
+    } else {
+      it++;
     }
-
-    iter = next;
   }
+  /* *INDENT-ON* */
+
+  GST_LOG_OBJECT (compositor, "Final queue size %" G_GSIZE_FORMAT,
+      priv->overlays.size ());
 
   return TRUE;
 }
@@ -633,41 +565,60 @@ gst_d3d11_overlay_compositor_draw_unlocked (GstD3D11OverlayCompositor *
     compositor, ID3D11RenderTargetView * rtv[GST_VIDEO_MAX_PLANES])
 {
   GstD3D11OverlayCompositorPrivate *priv;
-  GList *iter;
   ID3D11DeviceContext *context;
   ID3D11ShaderResourceView *clear_view[GST_VIDEO_MAX_PLANES] = { nullptr, };
   UINT strides = sizeof (VertexData);
   UINT offsets = 0;
+  ID3D11SamplerState *samplers[1];
 
   g_return_val_if_fail (compositor != nullptr, FALSE);
   g_return_val_if_fail (rtv != nullptr, FALSE);
 
   priv = compositor->priv;
 
-  if (!priv->overlays)
+  if (priv->overlays.empty ())
     return TRUE;
 
+  samplers[0] = priv->sampler.Get ();
+
   context = gst_d3d11_device_get_device_context_handle (compositor->device);
   context->IASetPrimitiveTopology (D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST);
-  context->IASetInputLayout (priv->layout);
-  context->IASetIndexBuffer (priv->index_buffer, DXGI_FORMAT_R16_UINT, 0);
-  context->PSSetSamplers (0, 1, &priv->sampler);
-  context->VSSetShader (priv->vs, nullptr, 0);
-  context->PSSetShader (priv->ps, nullptr, 0);
+  context->IASetInputLayout (priv->layout.Get ());
+  context->IASetIndexBuffer (priv->index_buffer.Get (),
+      DXGI_FORMAT_R16_UINT, 0);
+  context->PSSetSamplers (0, 1, samplers);
+  context->VSSetShader (priv->vs.Get (), nullptr, 0);
   context->RSSetViewports (1, &priv->viewport);
+  context->RSSetState (priv->rs.Get ());
   context->OMSetRenderTargets (1, rtv, nullptr);
-  context->OMSetBlendState (priv->blend, nullptr, 0xffffffff);
+  context->OMSetBlendState (priv->blend.Get (), nullptr, 0xffffffff);
 
-  for (iter = priv->overlays; iter; iter = g_list_next (iter)) {
-    GstD3D11CompositionOverlay *overlay =
-        (GstD3D11CompositionOverlay *) iter->data;
+  /* *INDENT-OFF* */
+  for (auto overlay : priv->overlays) {
+    ID3D11ShaderResourceView *srv[] = { overlay->srv.Get () };
+    ID3D11Buffer *vertex_buf[] = { overlay->vertex_buffer.Get () };
+    GstMapInfo info;
+    GstMemory *mem = nullptr;
 
-    context->PSSetShaderResources (0, 1, &overlay->srv);
-    context->IASetVertexBuffers (0,
-        1, &overlay->vertex_buffer, &strides, &offsets);
+    if (overlay->d3d11_buffer) {
+      mem = gst_buffer_peek_memory (overlay->d3d11_buffer, 0);
+      gst_memory_map (mem, &info, (GstMapFlags) (GST_MAP_D3D11 | GST_MAP_READ));
+    }
+
+    if (!overlay->premul_alpha)
+      context->PSSetShader (priv->premul_ps.Get (), nullptr, 0);
+    else
+      context->PSSetShader (priv->ps.Get (), nullptr, 0);
+
+    context->PSSetShaderResources (0, 1, srv);
+    context->IASetVertexBuffers (0, 1, vertex_buf, &strides, &offsets);
 
     context->DrawIndexed (6, 0, 0);
+
+    if (mem)
+      gst_memory_unmap (mem, &info);
   }
+  /* *INDENT-ON* */
 
   context->PSSetShaderResources (0, 1, clear_view);
   context->OMSetRenderTargets (0, nullptr, nullptr);
